#coding=utf-8
"""
original AMIS - reweighting of particles from previous iterations (populations) is done at each iteration step
"""

import copy
import gc
import time
import scipy.io
import scipy.stats
import math
import os
import numpy

import data.nuclideDB as ndb

try: #fast cython version
    from puff_cython.puff_tools import Puff
    from puff_cython import ppe 
    print "Cython version imported"   
except: #slow python version
    from puff_python.puff_tools import Puff
    from puff_python import ppe
    print "Python version imported"

def get_meteo(METEO, grid_step, t, X, Y):
    """
    interpolation of meteorological conditions in time and space
    """
    t0 = int(t / 60.0 / 60.0)

    x = X % grid_step
    y = Y % grid_step

    wd11_0 = METEO[t0, 0, 0, 0]
    wd12_0 = METEO[t0, 0, 0, 0]
    wd21_0 = METEO[t0, 0, 0, 0]
    wd22_0 = METEO[t0, 0, 0, 0]

    ws11_0 = METEO[t0, 1, 0, 0]
    ws12_0 = METEO[t0, 1, 0, 0]
    ws21_0 = METEO[t0, 1, 0, 0]
    ws22_0 = METEO[t0, 1, 0, 0]

    wd11_1 = METEO[t0+1, 0, 0, 0]
    wd12_1 = METEO[t0+1, 0, 0, 0]
    wd21_1 = METEO[t0+1, 0, 0, 0]
    wd22_1 = METEO[t0+1, 0, 0, 0]

    ws11_1 = METEO[t0+1, 1, 0, 0]
    ws12_1 = METEO[t0+1, 1, 0, 0]
    ws21_1 = METEO[t0+1, 1, 0, 0]
    ws22_1 = METEO[t0+1, 1, 0, 0]

    wd_0 = bilinear_interp(wd11_0, wd12_0, wd21_0, wd22_0, 0, 0, grid_step, grid_step, x, y)
    ws_0 = bilinear_interp(ws11_0, ws12_0, ws21_0, ws22_0, 0, 0, grid_step, grid_step, x, y)

    wd_1 = bilinear_interp(wd11_1, wd12_1, wd21_1, wd22_1, 0, 0, grid_step, grid_step, x, y)
    ws_1 = bilinear_interp(ws11_1, ws12_1, ws21_1, ws22_1, 0, 0, grid_step, grid_step, x, y)

    wd = (t-t0*60*60)/(60*60*(t0+1 - t0))*(wd_1-wd_0) + wd_0
    ws = (t-t0*60*60)/(60*60*(t0+1 - t0))*(ws_1-ws_0) + ws_0
    return (wd, ws)

####################################################################################################################
def bilinear_interp(q11, q12, q21, q22, x1, y1, x2, y2, x, y):
    """
    bilinear interpolation on a rectangular network
    """
    return q11*(x2-x)*(y2-y)/(x2-x1)/(y2-y1)+q21*(x-x1)*(y2-y)/(x2-x1)/(y2-y1)+q12*(x2-x)*(y-y1)/(x2-x1)/(y2-y1)+q22*(x-x1)*(y-y1)/(x2-x1)/(y2-y1)

####################################################################################################################
def loadReceptors(path):
    """
    list of receptors locations
    """
    
    X = []
    Y = []
    names = []
    f = open(path, "r")
    s = f.readlines()
    InputTuple = []
    for line in s[0:len(s)]:
        singleLine = line.rstrip()
        InputTuple.append(singleLine)
        name = singleLine[14:42]
        north = float(singleLine[42:54])
        east = float(singleLine[54:62])
        Y.append(float(east))
        X.append(float(north))
        names.append(name)
    f.close()
    return [X, Y, names]


####################################################################################################################
def eval_normal_ll(x, y, y_err):
    """
    normal likelihood
    """
    oC = y.shape[0] #observations count
    total_ll = 0.0
    C1 = 0.2
    K1 = BACKGROUND_DOSE
    for i in range(oC):
        delta = (x[i]+K1 - y[i])
        S = (C1 * (x[i]+K1))**2
        
        total_ll += -0.5 * ( math.log(S, math.e) + (delta**2)/(S))

    return total_ll

####################################################################################################################
def evaluate_weights(dose_of_models, t, Y, Y_err, LL, receptors_count, batch):
    """
    calculation of weights of particles in N-R optimization procedure
    """
    log_likelihoods = numpy.zeros(batch)
    
    for i in range(batch):
        X = numpy.sum(dose_of_models[i,t,:,:], axis=1)
        log_likelihoods[i] = eval_normal_ll( X, Y, Y_err ) + LL[i]
            
    VAHY = numpy.exp(log_likelihoods - log_likelihoods.max())
    VAHY = VAHY / numpy.sum(VAHY)
    
    return VAHY

####################################################################################################################
def evaluate_obs_ll(dose_of_models, t, Y, Y_err, receptors_count, batch): #computes numerator of logarithmic w
    """
    observation likelihood
    """
    log_likelihoods = numpy.zeros(batch)
    oC = Y.shape[0]
    
    X = numpy.sum(dose_of_models[:,t,:,:], axis=2)
    
    for i in range(batch):
        X = numpy.sum(dose_of_models[i,t,:,:], axis=1)
        log_likelihoods[i] = eval_normal_ll( X, Y, Y_err )
        
    return log_likelihoods
  
####################################################################################################################  
def evaluate_ll_denominator(pop, batch, models, KERNEL_MU, KERNEL_VAR, t, pN):
    """
    evaluates q in w=p/q with respect to given gaussian kernel 
    """
    denominators = numpy.zeros(models.shape[0])
    
    
    for i in range((pop+1)*batch):
        jacobians = models[i][0].wind_a
        new_vec = []  
        new_vec.append(math.log(models[i][0].wind_a))
        new_vec.append(models[i][0].wind_b)
        
        if (t < pN):
            jacobians *= models[i][t].Q  
            new_vec.append(math.log(models[i][t].Q))  
        
        new_vec = numpy.array(new_vec)
            
        for p in range(pop+1):
            mu_hat = KERNEL_MU[p]
            sigma_hat = KERNEL_VAR[p]
            llp = log_multi_gauss(numpy.reshape(new_vec, (mu_hat.shape[0],1)), mu_hat, sigma_hat) 
            denominators[i] += batch * (math.exp(llp) / jacobians)
    
           
    denominators = denominators / ((pop+1.)*batch)
     
    return numpy.log(denominators)

####################################################################################################################
def multinomial_resampling(VAHY, new_ensemble_size):
    """
    multinomial resampling - does not perform resampling itself, just generates numbers of particular particles in new sample
    according to their likelihoods
    """
    scn = new_ensemble_size #samples count
    sc = len(VAHY)
    BASKETS = numpy.zeros(sc)
    CDF_vah = numpy.zeros(sc+1)
    for i in range(1, sc):
        CDF_vah[i] = VAHY[i-1] + CDF_vah[i-1]
    CDF_vah[sc] = 1.
    U = numpy.sort(numpy.random.uniform(0., 1., scn))
    last_index = 0
    for i in range(scn):
        for j in range(last_index, sc):
            if (U[i] > CDF_vah[j] and U[i] <= CDF_vah[j+1] ):
                BASKETS[j] += 1
                last_index = j
                break
    return BASKETS

####################################################################################################################
def resample(models, cetnosti, RES_PAM, new_sample_size):
    """
    performs resampling using baskets generated using previsou method
    """
    res_models = numpy.zeros((new_sample_size, models.shape[1]), dtype=Puff)
    RES_PAM_NEW = numpy.zeros(RES_PAM.shape)
    index = 0
    for i in range(len(cetnosti)):
        for j in range(int(cetnosti[i])):
            for k in range(len(models[i])):
                puff = Puff(models[i][k].id, models[i][k].Q, models[i][k].wind_a, models[i][k].wind_b, models[i][k].Pcategory,
                            models[i][k].wind_c, models[i][k].x0, models[i][k].y0, models[i][k].H, models[i][k].tot_d, models[i][k].tot_t,
                            models[i][k].dd_total, models[i][k].rd_total, models[i][k].sigmaxy, models[i][k].sigmaz, models[i][k].lambda_decay,models[i][k].vG)
                res_models[index][k] = puff
            #resampling of RESULTS_PARAMS
            RES_PAM_NEW[:, index, :] = RES_PAM[:, i, :]
            index += 1
        
    VAHY = numpy.ones(new_sample_size) * 1.0 / new_sample_size
        
    return res_models, VAHY, RES_PAM_NEW

####################################################################################################################
def evol_b(b_old, sig_b, forecast, obs, sig_ywd):
    """
    statistical model of wind direction bias
    """
    r   = 1.0/(sig_b**(-2.0) + sig_ywd**(-2.0))
    mju = r * (sig_b**(-2.0)*b_old + sig_ywd**(-2.0)*(obs-forecast) )

    b_new = numpy.random.normal(mju, math.sqrt(r))
    
    q1 = log_normal(b_new, b_old, sig_b)  #likelihoood with respect to state evolution model 
    q2 = log_normal(b_new, mju, math.sqrt(r)) #likelihood with respecto to proposal
    q3 = log_normal(obs-forecast, b_new, sig_ywd) #likelihood with respect to observation
    
    ll = q1 - q2 + q3
    
    return b_new, ll, r, mju, q2

####################################################################################################################
def evol_a(a_old, gamm_a, forecast, obs, gamm_yws):
    """
    statistical model of wind speed bias
    """
    k  = gamm_a**(-2.0) + gamm_yws**(-2.0) + 2.0
    th = 1.0 / ( gamm_a**(-2.0)/a_old + forecast/obs * (gamm_yws**(-2)+1) )
    a_new = numpy.random.gamma(k, th) 
    
    q1 = log_gamma(a_new, k, th) #likelihood with respecto to proposal q(a_t|a_t-1, y_t)
    q2 = log_gamma(a_new, gamm_a**(-2), gamm_a**(2)*a_old) #likelihoood with respect to state evolution model - Gamma()
    q3 = log_gamma(forecast/obs, gamm_yws**(-2)+2, 1./ ((gamm_yws**(-2)+1)*a_new) ) #likelihood with respect to observation - iGamma(a) = Gamma(1/a)

    ll = q2 - q1 + q3

    return a_new, ll, k, th, q1

####################################################################################################################
def evol_Q(Q_old, Q_sigma):
    """
    transition model on Q
    """
    Q_new = numpy.random.gamma(Q_sigma**(-2.0), Q_sigma**2 * Q_old)
    return Q_new

####################################################################################################################
def compute_effective_number_of_particles(VAHY):
    """
    evaluation of effective number of particles
    """
    Neff =  1.0/numpy.sum(numpy.power(VAHY, 2))
    return Neff

####################################################################################################################
def log_gamma(x, k, th):
    """
    logarithm of gamma distribution
    """
    return (k-1)*math.log(x, math.e) - (x/th) - k*math.log(th, math.e) - math.lgamma(k)

####################################################################################################################
def log_normal(x, mju, sig): 
    """
    logarithm of normal distribution
    """
    return -1.0*math.log(sig, math.e) -0.5*((x-mju)/sig)**2 -0.5*math.log(2.*math.pi)

####################################################################################################################

class fdif:
    """
    class implementing first and second derivatives of logarithm of gamma pdf, see Appendix A
    """
    def __init__(self, alpha, beta, m, y, k, theta):
        self.alpha = alpha
        self.beta = beta
        self.m = m
        self.y = y
        self.k = k
        self.theta = theta

    def eval1(self, Q, Q_tran):
        ret = 0
        ret += numpy.sum(self.alpha*self.beta/(self.beta*Q+self.m) - self.beta/self.y)
        if (Q_tran):
            ret += (self.k-1)/Q     
            ret += - 1.0/self.theta 
        return ret 

    def eval2(self, Q, Q_tran):
        ret = -1.0* numpy.sum(self.alpha*self.beta**2/(self.beta*Q+self.m)**2 )
        if (Q_tran):
            ret -= (self.k-1)/Q**2
        return ret
    
####################################################################################################################
def Newton_Raphston(x0, fdif, Q_tran, thr=100): #thr - maximum number of iterations
    """
    method for finding a root of a well-behaved function:)
    here apllied to finding the root of the first derivative, i.e. finding a local minima of a function
    """
    x1 = numpy.Inf
    diff = numpy.Inf
    counter = 0
    
    xka = []
    fxka = []
    while (diff > 1.E-5 and counter < thr): #fdif is a stopping criterion
        x1 = x0 - fdif.eval1(x0, Q_tran)/fdif.eval2(x0, Q_tran)
        diff = abs(x0-x1)
        x0 = x1
        if (x0 < 0):
            x0=10
        xka.append(x0)
        fxka.append(fdif.eval1(x0, Q_tran))
        counter += 1

    return x0, math.sqrt(-1.0/2./fdif.eval2(x0, Q_tran))

####################################################################################################################
def get_stats(x, batch, VAHY):
    """
    computes mean and variance of particles' parameters
    """
    mu_hat = 0.
    sigma_hat = 0.
    for j in range(batch):
        x0 = numpy.reshape(x[j,:], (x.shape[1],1))
        mu_hat += VAHY[j]*x0
        sigma_hat += VAHY[j]*numpy.dot(x0, x0.transpose())             
    
    mumuT = numpy.dot(mu_hat, mu_hat.transpose())
    sigma_hat -= mumuT
    
    return mu_hat, sigma_hat, mumuT

####################################################################################################################
def get_params_of_particles(models_aux, batch, t):
    """
    extracts params (wind_a, wind_b, Q) from partciles
    """
    pN = models_aux.shape[1]
    x = 0.
    if (t < pN):
        x = numpy.zeros((batch, 3))
        for j in range(batch):   
            x[j,0] = math.log(models_aux[j][0].wind_a)
            x[j,1] = models_aux[j][0].wind_b
            x[j,2] = math.log(models_aux[j][t].Q)
    else: #we don't estimate Q anymore, the release is over
        x = numpy.zeros((batch, 2))
        for j in range(batch):
            x[j,0] = math.log(models_aux[j][0].wind_a)
            x[j,1] = models_aux[j][0].wind_b
            
    return x

####################################################################################################################
def log_multi_gauss(x, mu, sigma):
    """
    logarithm of multidomensional Gaussian distribution
    """
    delta = x-mu
    sigma_i = numpy.linalg.inv(sigma)
    sigma_det = numpy.linalg.det(sigma)
    k = mu.shape[0]
    
    return -0.5* ( k*math.log(2.*math.pi) + math.log(sigma_det) + numpy.dot( numpy.dot(delta.transpose(), sigma_i), delta) )

####################################################################################################################
def main(gamm_a, sig_b, gamm_yws, sig_ywd, gamm_Q, gamm_y, p_time, r_time, particles_count, batch, populC, input_data_path, METEO_OBS_DICT, n, Q_tran, kulhavy):
    propagation_time = p_time #total time of propagation of the model in sec
    rel_time = r_time #total time of release - it will be split into puffs according to dose_integration_time; in sec
    dose_integration_time = 10.*60. #a time interval of the release which is approximated with a single puff 
    dose_sub_step_no = 5 #the number of integration steps used in step-wise integration of dose with dose_integration_time
    
    total_steps = propagation_time / dose_integration_time #the total number of time steps 
    pN = int(rel_time / dose_integration_time) #number of puffs in a puff model calculated from dose_integration_time and the duration of release
    
    print "Initializing puff model with %d %d-minute puffs" % (pN, int(dose_integration_time/60.))
    print "Total steps: %d" % total_steps

    nuclideDB = ndb.NuclideDB()  #a simple database of radionuclides and their properties
    
    PUFF_PARAMS = {}
    PUFF_PARAMS["relH"] = 50 # height of release in meters
    PUFF_PARAMS["nucl_name"] = "Ar-41"
    PUFF_PARAMS["dose_integration_time"] = dose_integration_time #je to vlastne time_step length
    PUFF_PARAMS["dose_sub_step_no"] = dose_sub_step_no
    
    PARAMS = {}
    PARAMS["dX"] = 41 #dimension of the grid in x; odd number in order to make the grid symmetric with respect to the middle grid point
    PARAMS["dY"] = 41 #dimension of the grid in y;
    PARAMS["dS"] = 1.0 #step of the computational grid in km
    PARAMS["puff_number"]     = pN 
    PARAMS["total_steps"]     = total_steps
    PARAMS["nuclideDB"]       = nuclideDB
    PARAMS["PUFF_PARAMS"] = PUFF_PARAMS 

    
###########radiological and other physical parameters of the realsed
    H = PARAMS["PUFF_PARAMS"]["relH"]
    vG = nuclideDB.getDepoVG(PARAMS["PUFF_PARAMS"]["nucl_name"])
    lambda_decay = nuclideDB.getLambda(PARAMS["PUFF_PARAMS"]["nucl_name"])
    doseA     = nuclideDB.getDoseA(PARAMS["PUFF_PARAMS"]["nucl_name"])
    doseB     = nuclideDB.getDoseB(PARAMS["PUFF_PARAMS"]["nucl_name"])
    doseMJU   = nuclideDB.getDoseMJU(PARAMS["PUFF_PARAMS"]["nucl_name"])
    doseMJUaE = nuclideDB.getDoseMJUaE(PARAMS["PUFF_PARAMS"]["nucl_name"])
    doseCb    = nuclideDB.getDoseCb(PARAMS["PUFF_PARAMS"]["nucl_name"])


    #some initial setting
    Q0 = 100.0
    a0 = 1. 
    b0 = 0. 
    sc0 = 4  
    wb0 = 0.     

    
    #first batch is for adaptaion using NR
    models_NR = numpy.zeros((batch, pN), dtype=Puff) #initialization of models for prpopsal adapttaion
    for i in range(batch):
        for j in range(pN):
            puff = Puff(id=j,
                Q=Q0, wind_a=a0, wind_b=b0, Pcategory=sc0, wind_c=wb0, 
                x0=0., y0=0., H=H, tot_d=1., tot_t=1., dd_total=1., rd_total=1., sigmaxy=0.,
                sigmaz=0., lambda_decay=lambda_decay, vG=vG)
            models_NR[i,j] = puff
    
     
    models = numpy.zeros((particles_count-batch, pN), dtype=Puff)        
    for i in range(particles_count - batch):
        for j in range(pN):
            puff = Puff(id=j,
                Q=Q0, wind_a=a0, wind_b=b0, Pcategory=sc0, wind_c=wb0, 
                x0=0., y0=0., H=H, tot_d=1., tot_t=1., dd_total=1., rd_total=1., sigmaxy=0.,
                sigmaz=0., lambda_decay=lambda_decay, vG=vG)
            models[i,j] = puff            

################################RECEPTORS
    receptor_path = input_data_path+os.sep+"receptory_ETE"
    receptors = []
    
    X, Y, names = loadReceptors(receptor_path)
    for i in range(len(X)):
        receptors.append((X[i], Y[i], 0.0))
        
    receptors_count = len(X) #number of receptors    

    #cython needs nd_array
    receptors = numpy.array(receptors)
#########################################
    #measurements
    path_mea = input_data_path+os.sep+"measurements"
    meaDict = scipy.io.loadmat(path_mea)
    Y0 = meaDict["y"]
    Y_err = meaDict["y_err"]

##########################################
    #gridded numerical meteoforecast
    METEO = scipy.io.loadmat(input_data_path+os.sep+"meteo")["METEO"]
    
    #measured meteodata
    METEO_OBS_WS = METEO_OBS_DICT["ws"][:,0]      
    METEO_OBS_WD = METEO_OBS_DICT["wd"][:,0]

    #True - uses point meteoforecast
    #False - uses gridded meteoforecast from meteo.mat for forcing the model
    obs_meteo_flag = False

##########################################################

    start = time.time()

##########################################################
    dose_of_models_adapt = numpy.zeros((batch, int(propagation_time/dose_integration_time), len(receptors), pN))
    
    RESULTS_PARAMS = numpy.zeros((int(propagation_time/dose_integration_time), particles_count - batch, 4)) #0 - weight, 1-3 - params
    Neff = []
    WS_PROPOSAL = []                                                
    WD_PROPOSAL = []
        
    start = time.time()
    
    #VAHY_nnorm = p(y_t|x_t)p(x_t|x_{t-1})
    llWEIGHTS = numpy.zeros(particles_count - batch, dtype=numpy.double) #non-normalized log-weights
    llOBS = numpy.zeros(particles_count - batch, dtype=numpy.double) #non-normalized likelihood of observations
    llNUMERATOR = numpy.zeros(particles_count - batch, dtype=numpy.double) #non-normalized numerator of log-weights
    llDENOMINATOR = numpy.zeros(particles_count - batch, dtype=numpy.double) #non-normalized denominators of log-weights
    
    WEIGHTS = numpy.ones(particles_count - batch, dtype=numpy.double) / (particles_count - batch)
        
    Qs_after_resampling = numpy.zeros((int(propagation_time/dose_integration_time), particles_count - batch, pN))
    
    t_max = int(propagation_time/dose_integration_time)
    
    for t in range(t_max): #loop over time steps
        total_time_global = t * dose_integration_time
        print "TIME STEP %d, time=%3.2fmin" % (t, total_time_global/60.0)

        k0, th0 = 0., 0.
        mju0, r0 = 0., 0.
        
        #obtaining mesurements for the current time step
        y = Y0[t]
        y_err = Y_err[t]
        
        LL_adapt = numpy.zeros(particles_count) 
                
        POP_MU = numpy.zeros((populC+1, 3))
        POP_SIGMA = numpy.zeros((populC+1, 3))
        KERNEL_MU = []
        KERNEL_VAR = []   
        
        llws_adapt = numpy.zeros(batch)
        llwd_adapt = numpy.zeros(batch)
        NR_props = numpy.zeros((batch, 3))
        
        #the first batch - mandatory adaptation step with Newton Rapson
        for j in range(batch): #sampling from new proposal on meteorological condition - wind speed and direction
            wd_fcst, ws_fcst = get_meteo(METEO, 3.0, int(total_time_global / 60.0 / 60.0), models_NR[j][0].x0/1000.0, models_NR[j][0].y0/1000.0)
            new_ws, llws_adapt[j], k0, th0, lprop_a = evol_a(models_NR[j][0].wind_a, gamm_a,  ws_fcst, METEO_OBS_WS[t], gamm_yws)
            new_wd, llwd_adapt[j], r0, mju0, lprop_b = evol_b(models_NR[j][0].wind_b, sig_b, wd_fcst, METEO_OBS_WD[t], sig_ywd)
            
            LL_adapt[j] = llws_adapt[j] + llwd_adapt[j]
            
            NR_props[j][0] = lprop_a 
            NR_props[j][1] = lprop_b
            
            for k in range(len(models_NR[j])): #new samples of meteorological conditions are assigned to particles
                models_NR[j][k].wind_a = new_ws
                models_NR[j][k].wind_b = new_wd

        #logging
        WS_PROPOSAL.append((k0, th0)) 
        WD_PROPOSAL.append((mju0, r0)) 

        print "Waiting to finish all jobs..."
        (models_NR, ci_of_models_adapt, dose_of_models_adapt[:, t, :, :]) = ppe.parallel_puff(j, total_time_global, models_NR, receptors, doseA, doseB, doseMJU, doseMJUaE, doseCb, dose_integration_time, dose_sub_step_no, n, METEO, obs_meteo_flag, METEO_OBS_WS, METEO_OBS_WD)
        ci = numpy.array(ci_of_models_adapt)
        print "Finished!"
        
        #evaluation of weights and resampling
        for j in range(batch):
            if (t < pN):
                Q_old = 0
                if (t == 0):
                    Q_old = models_NR[j][t].Q
                else:
                    Q_old = models_NR[j][t-1].Q

                alpha = gamm_y**(-2)+2
                beta = (alpha-1)*ci[j,:receptors_count,t] 
                m = (alpha-1)*(numpy.sum(dose_of_models_adapt[j, t, :receptors_count, 0:t], axis=1)+BACKGROUND_DOSE)
                k = gamm_Q**(-2.)
                theta = gamm_Q**2 * Q_old
                
                fdif0 = fdif(alpha, beta, m, y, k, theta)
                Q_max, Q_std = Newton_Raphston(Q_old, fdif0, Q_tran, thr=500)
                Q_new = scipy.stats.truncnorm.rvs(-1.0*Q_max/Q_std, numpy.Inf, loc=Q_max, scale=Q_std)
                    
                models_NR[j][t].Q = Q_new
                LL_adapt[j] +=  - log_normal(Q_new, Q_max, Q_std) 
                
                if (Q_tran):
                    LL_adapt[j] +=  log_gamma(Q_new, gamm_Q**(-2.0), gamm_Q**2.0 * Q_old) 
                
                dose_of_models_adapt[j, t, :, t] = Q_new * ci[j,:,t] 
                NR_props[j][2] = log_normal(Q_new, Q_max, Q_std)
                
        print "time=", t, "Newton Raphson"
                
        #the first population was done using N-R with UNIFORM weights
        x = numpy.zeros((batch, 3))
        x = get_params_of_particles(models_NR, batch, t)
        mu_hat, sigma_hat, mumuT = get_stats(x, batch, numpy.ones(batch)*1./batch) #calcualtion of mu_hat and sigma_hat using (36)
        
        #the first Gaussian proposal
        print "After initial Newton-Rapson procedure with %d particles (uniform weights): " % batch
        print "mu_hat:"
        print mu_hat.transpose()
        print "Sigma_hat:"
        print numpy.sqrt(numpy.diag(sigma_hat))
       
        #first proposal kernel estimated by NR 
        if (t < pN):
            POP_MU[0,:] = mu_hat[:,0]
            POP_SIGMA[0,:] =  numpy.sqrt(numpy.diag(sigma_hat))[:] 
        else:
            POP_MU[0,:2] = mu_hat[:,0]
            POP_SIGMA[0,:2] =  numpy.sqrt(numpy.diag(sigma_hat))[:]

        #kernels of adaptive proposals for first step 
        KERNEL_MU.append(mu_hat)
        KERNEL_VAR.append(sigma_hat)   
                
        #inital values 
        gamm_0 = 100.
        nu_0 = 1.
        x_hat_t0 = mu_hat
        lambda_0 = sigma_hat
        
        #models_adapt stay unchanged in adaptation steps since its propagated state is discarded after trial progation
        VAHY_adapt_print = evaluate_weights(dose_of_models_adapt, t, y, y_err, LL_adapt, receptors_count, batch)
          
        print "===neff after N-R:", compute_effective_number_of_particles(VAHY_adapt_print)
###############ITERATION OF POPULATIONS
        # we split the total amount of particles 1000 into populations of 100 and iterate 10 times
        st = time.time()
        for p in range(populC): 
            
            for j in range(batch):
                wd_fcst, ws_fcst = get_meteo(METEO, 3.0, int(total_time_global / 60.0 / 60.0), models[p*batch+j][0].x0/1000.0, models[p*batch+j][0].y0/1000.0)
                #sampling of new values for current population
                new_vec = numpy.random.multivariate_normal(mu_hat[:,0], sigma_hat)
                
                new_a0 = math.exp(new_vec[0])
                llws0 = log_gamma(new_a0, gamm_a**(-2), gamm_a**(2)*models[p*batch+j][0].wind_a) #likelihood of state transition wind speed (WS)
                llws0 += log_gamma(ws_fcst/METEO_OBS_WS[t], gamm_yws**(-2)+2, 1./ ((gamm_yws**(-2)+1)*new_a0) ) #observation likelihood WS
                      
                llwd0 = log_normal(new_vec[1], models[p*batch+j][0].wind_b, sig_b)  #likelihood of state transition wind direction (WD)
                llwd0 += log_normal(METEO_OBS_WD[t]-wd_fcst, new_vec[1], sig_ywd) #observation likelihood WD
            
                llQ = 0. 
                
                if (t < pN):                    
                    if (Q_tran):
                        Q_new = math.exp(new_vec[2])
                        Q_old = models[p*batch+j][t].Q
                        llQ = log_gamma(Q_new, gamm_Q**(-2.0), gamm_Q**2.0 * Q_old)
                                  
                llNUMERATOR[p*batch+j] = llws0 + llwd0 + llQ

                for k in range(len(models[p*batch+j])): 
                    models[p*batch+j][k].wind_a = new_a0
                    models[p*batch+j][k].wind_b = new_vec[1]
                if (t < pN):
                    models[p*batch+j][t].Q = math.exp(new_vec[2])

                    
            (models[p*batch:(p+1)*batch], ci_of_models_sink, dose_of_models_adapt[:, t, :, :]) = ppe.parallel_puff(j, total_time_global, models[p*batch:(p+1)*batch], receptors, doseA, doseB, doseMJU, doseMJUaE, doseCb, dose_integration_time, dose_sub_step_no, n, METEO, obs_meteo_flag, METEO_OBS_WS, METEO_OBS_WD)
        
            #we evaluate denominators of weigth from current batch
            llOBS[p*batch:(p+1)*batch] = evaluate_obs_ll(dose_of_models_adapt, t, y, y_err, receptors_count, batch)             
            llDENOMINATOR =  evaluate_ll_denominator(p, batch, models[:(p+1)*batch], KERNEL_MU, KERNEL_VAR, t, pN) 
            llWEIGHTS = llOBS[:(p+1)*batch] + llNUMERATOR[:(p+1)*batch] - llDENOMINATOR[:(p+1)*batch]
            llWEIGHTS  = llWEIGHTS - llWEIGHTS.max()
            WEIGHTS = numpy.exp(llWEIGHTS) / numpy.exp(llWEIGHTS).sum()
            
            #obtaining effective number of particles
            neff = compute_effective_number_of_particles(WEIGHTS)
            print "===neff of population", p, ":", neff 
            
            #obtainit x - parameters of particles
            x = get_params_of_particles(models[:(p+1)*batch], (p+1)*batch, t)
            
            #obtaining mu_hat, sigma_hat - moments of Gaussian kernel
            mu_hat, sigma_hat, mumuT = get_stats(x, (p+1)*batch, WEIGHTS) #calcualtion of mu_hat and sigma_hat
            
            #updating of statistics using information geometry
            if (kulhavy):
                gamm_t = gamm_0 / (1 + gamm_0 * neff)
                x_hat_t = x_hat_t0 + gamm_t*neff*(mu_hat - x_hat_t0)
                nu_t = nu_0 + neff
                xt0xt0T = numpy.dot(x_hat_t0, x_hat_t0.transpose())
                xtxtT = numpy.dot(x_hat_t, x_hat_t.transpose())
                lambda_t = lambda_0 +  neff*( sigma_hat + mumuT - xtxtT) + 1./gamm_0 * (xt0xt0T - xtxtT) 
                            
                mu_hat = x_hat_t
                sigma_hat = lambda_t / nu_t 
            
            print "Proposal %d" % (p)
            print "mu_hat:"
            print mu_hat.transpose()
            print "Sigma_hat:"
            print numpy.sqrt(numpy.diag(sigma_hat))
            
        
            if (t < pN):
                POP_MU[p+1,:] = mu_hat[:,0]
                POP_SIGMA[p+1,:] =  numpy.sqrt(numpy.diag(sigma_hat))[:]
            else:
                POP_MU[p+1,:2] = mu_hat[:,0]
                POP_SIGMA[p+1,:2] =  numpy.sqrt(numpy.diag(sigma_hat))[:]

            KERNEL_MU.append(mu_hat)
            KERNEL_VAR.append(sigma_hat)       
        
        print "\n\n popul time:", time.time() - st     

        neff = compute_effective_number_of_particles(WEIGHTS)
        Neff.append( neff )
        print "Final effective number of particles =", neff  
            
            
        models, WEIGHTS, RESULTS_PARAMS = resample(models, multinomial_resampling(WEIGHTS, particles_count - batch), RESULTS_PARAMS, particles_count - batch)
        #we prepare small ensemble for adaptation in the next time step
        models_NR, sink, sink = resample(models, multinomial_resampling(numpy.ones(batch)*1./batch, batch), RESULTS_PARAMS, batch)
        
        #saving of params after resampling
        for j in range(particles_count - batch):                    
            if (t < pN):
                RESULTS_PARAMS[t, j, 1] = models[j][t].Q 
            RESULTS_PARAMS[t, j, 2] = models[j][0].wind_a 
            RESULTS_PARAMS[t, j, 3] = models[j][0].wind_b 
        
        for i in range(particles_count - batch):
            for j in range(pN):
                Qs_after_resampling[t, i, j] = models[i, j].Q

            
    #display execution time
    print "time: ", time.time() - start

    #saving of results
    dic = {}
    dic["results"] = RESULTS_PARAMS
    dic["Neff"] = Neff
    dic["WS_PROPOSAL"] = WS_PROPOSAL
    dic["WD_PROPOSAL"] = WD_PROPOSAL
    dic["METEO_OBS_WD"] = METEO_OBS_WD
    dic["METEO_OBS_WS"] = METEO_OBS_WS
    dic["dose_of_models"] = dose_of_models_adapt
    dic["gamm_Q"] = gamm_Q
    dic["gamm_y"] = gamm_y
    dic["gamm_a"] = gamm_a
    dic["sig_b"] = sig_b
    dic["mC"] = particles_count
    dic["Qs_after_resampling"] = Qs_after_resampling
            
    
    return dic    
    
if __name__ == "__main__":
    #path to input files
    input_data_path = "data"
    
    #path of output
    results_path = "results"+os.sep+"output"
    
    #file with meteorological observations
    METEO_OBS_DICT = scipy.io.loadmat("data"+os.sep+"meteo_obs")

    #total number of particles (adaptation particles + final population from best adapted kernel used for filtering)        
    particles_count = 1000 
    
    #number of particles used for adaptation in a SINGLE pupulation
    batch = 100 
    
    #number of adaptation steps AFTER NEWTON RAPHSON 
    populations = 9 
    
    #file with twin measurements (synthetic data)
    mea_dict = scipy.io.loadmat("data"+os.sep+"measurements")
    
    #total propagation time of the release
    propagation_time = 3.*60.*60. #total time of propagation of the model in sec
    
    #duration of the release
    release_time = mea_dict["release_time"] #1.*60.*60. #total time of release - it will be split into puffs according to dose_integration_time; in sec    
    print "Release time:", release_time
    
    #number of integration abscisas in Gauss-Legendre integration formulas used for evaluation of gamma dose from puffs
    n = mea_dict["n"] 
    
    #parameters of statistical models
    gamm_a = 0.2
    sig_b =  20.0 /360.0 * 2.0*math.pi
    gamm_Q = 1.0
    gamm_y = mea_dict["C"] 
    BACKGROUND_DOSE = mea_dict["K"] 
    gamm_v = METEO_OBS_DICT["gamm_yws"] #
    sig_phi  = METEO_OBS_DICT["sig_ywd"]
    
    #flag for transition model on Q
    Q_tran = False 
    
    #flag for information geometry
    kulhavy = True 
    
    #number of runs for purposes of MC study
    runs = 1 
    
    #label of output file
    label = "AMIS"
    
    for r in range(runs):
        dic = main(gamm_a, sig_b, gamm_v, sig_phi, gamm_Q, gamm_y, propagation_time, release_time, particles_count, batch, populations, input_data_path, METEO_OBS_DICT, n, Q_tran, kulhavy)
        
        #saving of results...
        scipy.io.savemat(results_path+os.sep+"results_%s_%d" % (label, r), dic, do_compression=True)
        print "Neff:", dic["Neff"]
    
    




